#include "MeshLoader.h"
using namespace std;

vector<Node>& MeshLoader::Get_Nodes() {
    return nodes;
}
vector<FiniteElement>& MeshLoader::Get_FiniteElements() {
    return finiteelements;
}
vector<BoundaryFiniteElement>& MeshLoader::Get_BFEs() {
    return bfes;
}


//!!! Все контейнеры должны возвращать ID, а не сами узлы и КЭ.

vector<FiniteElement> MeshLoader::Get_FE_by_nodes(int id1, int id2, int id3) {
    vector<FiniteElement> ans;
    auto it = finiteelements.begin();
    while (it != finiteelements.end()){
        it = find_if(it, finiteelements.end(), [id1, id2, id3](const FiniteElement &x) {
            return ((find(x.nodes_id.begin(), x.nodes_id.end(), id1) != x.nodes_id.end()) &&
                    (find(x.nodes_id.begin(), x.nodes_id.end(), id2) != x.nodes_id.end()) &&
                    (find(x.nodes_id.begin(), x.nodes_id.end(), id3) != x.nodes_id.end())
            );
        });
        if (it != finiteelements.end()) {
            ans.push_back(*it);
            it++;
        }
    }
    return ans;
}
vector<FiniteElement> MeshLoader::Get_FE_by_edge(int id1, int id2) {
    vector<FiniteElement> ans;
    auto it = finiteelements.begin();
    while (it != finiteelements.end()) {
        it = find_if(it, finiteelements.end(), [id1, id2](const FiniteElement &x)  {
            return ((find(x.nodes_id.begin(), x.nodes_id.end(), id1) != x.nodes_id.end()) &&
                    (find(x.nodes_id.begin(), x.nodes_id.end(), id2) != x.nodes_id.end())
            );
        });
        if (it != finiteelements.end()) {
            ans.push_back(*it);
            it++;
        }
    }
    return ans;
}

vector<Node> MeshLoader::Get_nodes_by_border(int id) {
    vector <Node> ans;
    auto it = bfes.begin();
    while(it != bfes.end()){
        it = find_if(it, bfes.end(), [id](BoundaryFiniteElement &x){ //!!! Почему нет const ?
            return x.border_id = id;});
        for (const auto& x : it->nodes_id)
            ans.push_back(nodes[x]);       //!!! Где гарантия, что узлы не будут дублироваться? Неверно выбран контейнер.
        it++;
    }
    return ans;
}
vector<FiniteElement>MeshLoader::Get_FE_by_material(int id) {
    vector<FiniteElement> ans;
    auto it = finiteelements.begin();
    while (it != finiteelements.end()) {
        it = find_if(it, finiteelements.end(), [id](const FiniteElement& x) {
            return x.mat_id == id; });
        if (it != finiteelements.end()) {
            ans.push_back(*it);
            it++;
        }
    }
    return ans;
}

vector<BoundaryFiniteElement> MeshLoader::Get_BFE_by_border(int id) {
    vector<BoundaryFiniteElement> ans;
    auto it = bfes.begin();
    while (it != bfes.end()) {
        it = find_if(it, bfes.end(), [id](const BoundaryFiniteElement &x) {
            return x.border_id == id; });
        if (it != bfes.end()) {
            ans.push_back(*it);
            it++;
        }
    }
    return ans;
}


const Node& MeshLoader::getnode(int x) {
    //!!! Крайне не эффективная реализация. Найти узел можно, не обходя все узлы, т.к. ID известен.
    
    return *find_if(nodes.begin(), nodes.end(), [x](const Node& n){return n.id == x;});
}

void MeshLoader::Insertion()
{
    //!!! Нужно завести хеш-таблицу для хеширования просмотренных ребер.
    //!!! Обходя КЭ и его ребра, нужно проверять, было ли просмотрено ребро ранее.
    //!!! Если да, то в ребре можно хранить ID третьего узла и использовать его для добавления в узлы данного КЭ.
    //!!! Если нет, то создать новый узел и захешировать ребро.
    
    vector<FiniteElement> inserted;
    for (auto &cur : finiteelements) {
        
        //!!! Крайне не эффективная реализация (сложность O(FE_COUNT ^ 2))
        
        if(find(inserted.begin(), inserted.end(), cur) == inserted.end()){
            int p = cur.nodes_id.size() / 2;
            int l = cur.nodes_id[p-1];
            int r = cur.nodes_id[p];
            Node node = {(int)nodes.size() + 1, (getnode(l).x + getnode(r).x)/2, (getnode(l).y + getnode(r).y)/2,(getnode(l).z + getnode(r).z)/2, false};
            cur.nodes_id.push_back(node.id);
            for (auto &x : finiteelements)
            {
                if (x != cur){
                    if((find(x.nodes_id.begin(), x.nodes_id.end(), l) != x.nodes_id.end()) && (find(x.nodes_id.begin(), x.nodes_id.end(), r) != x.nodes_id.end())){
                        inserted.push_back(x);
                        x.nodes_id.push_back(node.id);
                    }
                }
            }
            nodes.push_back(move(node));
        }
    }
    for (auto &cur : bfes) {
        if(find(inserted.begin(), inserted.end(), cur) == inserted.end()){
            int p = cur.nodes_id.size() / 2;
            int l = cur.nodes_id[p-1];
            int r = cur.nodes_id[p];
            Node node = {(int)nodes.size() + 1, (getnode(l).x + getnode(r).x)/2, (getnode(l).y + getnode(r).y)/2,(getnode(l).z + getnode(r).z)/2, false};
            cur.nodes_id.push_back(node.id);
            for (auto &x : bfes)
            {
                if (x != cur){
                    if((find(x.nodes_id.begin(), x.nodes_id.end(), l) != x.nodes_id.end()) && (find(x.nodes_id.begin(), x.nodes_id.end(), r) != x.nodes_id.end())){
                        inserted.push_back(x);
                        x.nodes_id.push_back(node.id);
                    }
                }
            }
            nodes.push_back(move(node));
        }
    }
}

vector <set <int>> MeshLoader::Get_neighbors(){
    vector<set<int>> ans(nodes.size() + 1);
    for (const auto &cur : finiteelements)
        for (const auto &id: cur.nodes_id)
            for (const auto &nbr: cur.nodes_id)
                if (id != nbr)
                    ans[id].insert(nbr);
    return ans;
}

void MeshLoader::printNode(const Node &node) {
    std::cout << node;
}

void MeshLoader::printFE(const FiniteElement &fe) {
    std::cout << fe;
}

void MeshLoader::printBFE(const BoundaryFiniteElement &bfe) {
    std::cout << bfe;
}

vector<int> MeshLoader::getBoundaryNodesId() {
    std::vector<int> res;
    for (const auto &it: bfes) { //!!! it - это не итератор. Переименуйте.
        std::for_each(it.nodes_id.begin(), it.nodes_id.end(), [&res](int id) {
            
            res.push_back(id); //!!! Где гарантия, что узлы не будут дублироваться?
        });
    }
    return res;
}


